@@ -2,7 +2,11 @@
 
 import copy
 import platform
-import unittest
+import sys
+if sys.version_info < (2, 7):
+    import unittest2 as unittest
+else:
+    import unittest
 
 from decimal import Decimal
 
@@ -10,10 +14,22 @@ from pymysqlreplication.tests import base
 from pymysqlreplication.constants.BINLOG import *
 from pymysqlreplication.row_event import *
 from pymysqlreplication.event import *
+from pymysqlreplication._compat import text_type
+
 
 __all__ = ["TestDataType"]
 
 
+def to_binary_dict(d):
+    def encode_value(v):
+        if isinstance(v, text_type):
+            return v.encode()
+        if isinstance(v, list):
+            return [encode_value(x) for x in v]
+        return v
+    return dict([(k.encode(), encode_value(v)) for (k, v) in d.items()])
+
+
 class TestDataType(base.PyMySQLReplicationTestCase):
     def ignoredEvents(self):
         return [GtidEvent]
@@ -153,6 +169,12 @@ class TestDataType(base.PyMySQLReplicationTestCase):
         self.assertEqual(type(event.rows[0]["values"]["test"]), type(None))
         self.assertEqual(event.rows[0]["values"]["test"], None)
 
+    def test_tiny_maps_to_none_2(self):
+        create_query = "CREATE TABLE test (test BOOLEAN)"
+        insert_query = "INSERT INTO test VALUES(NULL)"
+        event = self.create_and_insert_value(create_query, insert_query)
+        self.assertEqual(event.rows[0]["values"]["test"], None)
+
     def test_short(self):
         create_query = "CREATE TABLE test (id SMALLINT UNSIGNED NOT NULL, test SMALLINT)"
         insert_query = "INSERT INTO test VALUES(65535, -32768)"
@@ -190,6 +212,7 @@ class TestDataType(base.PyMySQLReplicationTestCase):
     def test_timestamp_mysql56(self):
         if not self.isMySQL56AndMore():
             self.skipTest("Not supported in this version of MySQL")
+        self.set_sql_mode()
         create_query = '''CREATE TABLE test (test0 TIMESTAMP(0),
             test1 TIMESTAMP(1),
             test2 TIMESTAMP(2),
@@ -206,11 +229,11 @@ class TestDataType(base.PyMySQLReplicationTestCase):
             '1984-12-03 12:33:07.123456')'''
         event = self.create_and_insert_value(create_query, insert_query)
         self.assertEqual(event.rows[0]["values"]["test0"], datetime.datetime(1984, 12, 3, 12, 33, 7))
-        self.assertEqual(event.rows[0]["values"]["test1"], datetime.datetime(1984, 12, 3, 12, 33, 7, 1))
-        self.assertEqual(event.rows[0]["values"]["test2"], datetime.datetime(1984, 12, 3, 12, 33, 7, 12))
-        self.assertEqual(event.rows[0]["values"]["test3"], datetime.datetime(1984, 12, 3, 12, 33, 7, 123))
-        self.assertEqual(event.rows[0]["values"]["test4"], datetime.datetime(1984, 12, 3, 12, 33, 7, 1234))
-        self.assertEqual(event.rows[0]["values"]["test5"], datetime.datetime(1984, 12, 3, 12, 33, 7, 12345))
+        self.assertEqual(event.rows[0]["values"]["test1"], datetime.datetime(1984, 12, 3, 12, 33, 7, 100000))
+        self.assertEqual(event.rows[0]["values"]["test2"], datetime.datetime(1984, 12, 3, 12, 33, 7, 120000))
+        self.assertEqual(event.rows[0]["values"]["test3"], datetime.datetime(1984, 12, 3, 12, 33, 7, 123000))
+        self.assertEqual(event.rows[0]["values"]["test4"], datetime.datetime(1984, 12, 3, 12, 33, 7, 123400))
+        self.assertEqual(event.rows[0]["values"]["test5"], datetime.datetime(1984, 12, 3, 12, 33, 7, 123450))
         self.assertEqual(event.rows[0]["values"]["test6"], datetime.datetime(1984, 12, 3, 12, 33, 7, 123456))
 
     def test_longlong(self):
@@ -244,17 +267,53 @@ class TestDataType(base.PyMySQLReplicationTestCase):
         self.assertEqual(event.rows[0]["values"]["test"], None)
         self.assertEqual(event.rows[0]["values"]["test2"], None)
 
-    def test_time(self):
-        create_query = "CREATE TABLE test (test TIME);"
-        insert_query = "INSERT INTO test VALUES('12:33:18')"
+    def test_zero_month(self):
+        self.set_sql_mode()
+        create_query = "CREATE TABLE test (id INTEGER, test DATE, test2 DATE);"
+        insert_query = "INSERT INTO test (id, test2) VALUES(1, '2015-00-21')"
         event = self.create_and_insert_value(create_query, insert_query)
-        self.assertEqual(event.rows[0]["values"]["test"], datetime.time(12, 33, 18))
+        self.assertEqual(event.rows[0]["values"]["test"], None)
+        self.assertEqual(event.rows[0]["values"]["test2"], None)
+
+    def test_zero_day(self):
+        self.set_sql_mode()
+        create_query = "CREATE TABLE test (id INTEGER, test DATE, test2 DATE);"
+        insert_query = "INSERT INTO test (id, test2) VALUES(1, '2015-05-00')"
+        event = self.create_and_insert_value(create_query, insert_query)
+        self.assertEqual(event.rows[0]["values"]["test"], None)
+        self.assertEqual(event.rows[0]["values"]["test2"], None)
+
+    def test_time(self):
+        create_query = "CREATE TABLE test (test1 TIME, test2 TIME);"
+        insert_query = "INSERT INTO test VALUES('838:59:59', '-838:59:59')"
+        event = self.create_and_insert_value(create_query, insert_query)
+        self.assertEqual(event.rows[0]["values"]["test1"], datetime.timedelta(
+            microseconds=(((838*60) + 59)*60 + 59)*1000000
+        ))
+        self.assertEqual(event.rows[0]["values"]["test2"], datetime.timedelta(
+            microseconds=(((-838*60) + 59)*60 + 59)*1000000
+        ))
+
+    def test_time2(self):
+        if not self.isMySQL56AndMore():
+            self.skipTest("Not supported in this version of MySQL")
+        create_query = "CREATE TABLE test (test1 TIME(6), test2 TIME(6));"
+        insert_query = """
+            INSERT INTO test VALUES('838:59:59.000000', '-838:59:59.000000');
+        """
+        event = self.create_and_insert_value(create_query, insert_query)
+        self.assertEqual(event.rows[0]["values"]["test1"], datetime.timedelta(
+            microseconds=(((838*60) + 59)*60 + 59)*1000000 + 0
+        ))
+        self.assertEqual(event.rows[0]["values"]["test2"], datetime.timedelta(
+            microseconds=(((-838*60) + 59)*60 + 59)*1000000 + 0
+        ))
 
     def test_zero_time(self):
         create_query = "CREATE TABLE test (id INTEGER, test TIME NOT NULL DEFAULT 0);"
         insert_query = "INSERT INTO test (id) VALUES(1)"
         event = self.create_and_insert_value(create_query, insert_query)
-        self.assertEqual(event.rows[0]["values"]["test"], datetime.time(0,0))
+        self.assertEqual(event.rows[0]["values"]["test"], datetime.timedelta(seconds=0))
 
     def test_datetime(self):
         create_query = "CREATE TABLE test (test DATETIME);"
@@ -263,18 +322,23 @@ class TestDataType(base.PyMySQLReplicationTestCase):
         self.assertEqual(event.rows[0]["values"]["test"], datetime.datetime(1984, 12, 3, 12, 33, 7))
 
     def test_zero_datetime(self):
+        self.set_sql_mode()
         create_query = "CREATE TABLE test (id INTEGER, test DATETIME NOT NULL DEFAULT 0);"
         insert_query = "INSERT INTO test (id) VALUES(1)"
         event = self.create_and_insert_value(create_query, insert_query)
         self.assertEqual(event.rows[0]["values"]["test"], None)
 
     def test_broken_datetime(self):
+        self.set_sql_mode()
         create_query = "CREATE TABLE test (test DATETIME NOT NULL);"
         insert_query = "INSERT INTO test VALUES('2013-00-00 00:00:00')"
         event = self.create_and_insert_value(create_query, insert_query)
         self.assertEqual(event.rows[0]["values"]["test"], None)
 
     def test_year(self):
+        if self.isMySQL57():
+            # https://dev.mysql.com/doc/refman/5.7/en/migrating-to-year4.html
+            self.skipTest("YEAR(2) is unsupported in mysql 5.7")
         create_query = "CREATE TABLE test (a YEAR(4), b YEAR(2))"
         insert_query = "INSERT INTO test VALUES(1984, 1984)"
         event = self.create_and_insert_value(create_query, insert_query)
@@ -320,6 +384,18 @@ class TestDataType(base.PyMySQLReplicationTestCase):
         self.assertEqual(event.rows[0]["values"]["test"], 'ba')
         self.assertEqual(event.rows[0]["values"]["test2"], 'a')
 
+    def test_enum_empty_string(self):
+        create_query = "CREATE TABLE test (test ENUM('a', 'ba', 'c'), test2 ENUM('a', 'ba', 'c')) CHARACTER SET latin1 COLLATE latin1_bin;"
+        insert_query = "INSERT INTO test VALUES('ba', 'asdf')"
+        last_sql_mode = self.execute("SELECT @@SESSION.sql_mode;"). \
+            fetchall()[0][0]
+        self.execute("SET SESSION sql_mode = 'ANSI';")
+        event = self.create_and_insert_value(create_query, insert_query)
+        self.execute("SET SESSION sql_mode = '%s';" % last_sql_mode)
+
+        self.assertEqual(event.rows[0]["values"]["test"], 'ba')
+        self.assertEqual(event.rows[0]["values"]["test2"], '')
+
     def test_set(self):
         create_query = "CREATE TABLE test (test SET('a', 'ba', 'c'), test2 SET('a', 'ba', 'c')) CHARACTER SET latin1 COLLATE latin1_bin;"
         insert_query = "INSERT INTO test VALUES('ba,a,c', 'a,c')"
@@ -367,6 +443,114 @@ class TestDataType(base.PyMySQLReplicationTestCase):
         event = self.create_and_insert_value(create_query, insert_query)
         self.assertEqual(event.rows[0]["values"]["test"], b'\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\xf0?')
 
+    def test_json(self):
+        if not self.isMySQL57():
+            self.skipTest("Json is only supported in mysql 5.7")
+        create_query = "CREATE TABLE test (id int, value json);"
+        insert_query = """INSERT INTO test (id, value) VALUES (1, '{"my_key": "my_val", "my_key2": "my_val2"}');"""
+        event = self.create_and_insert_value(create_query, insert_query)
+        self.assertEqual(event.rows[0]["values"]["value"], {b"my_key": b"my_val", b"my_key2": b"my_val2"})
+
+    def test_json_array(self):
+        if not self.isMySQL57():
+            self.skipTest("Json is only supported in mysql 5.7")
+        create_query = "CREATE TABLE test (id int, value json);"
+        insert_query = """INSERT INTO test (id, value) VALUES (1, '["my_val", "my_val2"]');"""
+        event = self.create_and_insert_value(create_query, insert_query)
+        self.assertEqual(event.rows[0]["values"]["value"], [b'my_val', b'my_val2'])
+
+    def test_json_large(self):
+        if not self.isMySQL57():
+            self.skipTest("Json is only supported in mysql 5.7")
+        data = dict([('foooo%i'%i, 'baaaaar%i'%i) for i in range(2560)])  # Make it large enough to reach 2^16 length
+        create_query = "CREATE TABLE test (id int, value json);"
+        insert_query = """INSERT INTO test (id, value) VALUES (1, '%s');""" % json.dumps(data)
+        event = self.create_and_insert_value(create_query, insert_query)
+
+        self.assertEqual(event.rows[0]["values"]["value"], to_binary_dict(data))
+
+    def test_json_large_with_literal(self):
+        if not self.isMySQL57():
+            self.skipTest("Json is only supported in mysql 5.7")
+        data = dict([('foooo%i'%i, 'baaaaar%i'%i) for i in range(2560)], literal=True)  # Make it large with literal
+        create_query = "CREATE TABLE test (id int, value json);"
+        insert_query = """INSERT INTO test (id, value) VALUES (1, '%s');""" % json.dumps(data)
+        event = self.create_and_insert_value(create_query, insert_query)
+
+        self.assertEqual(event.rows[0]["values"]["value"], to_binary_dict(data))
+
+    def test_json_types(self):
+        if not self.isMySQL57():
+            self.skipTest("Json is only supported in mysql 5.7")
+
+        types = [
+           True,
+           False,
+           None,
+           1.2,
+           2^14,
+           2^30,
+           2^62,
+           -1 * 2^14,
+           -1 * 2^30,
+           -1 * 2^62,
+           ['foo', 'bar']
+        ]
+
+        for t in types:
+            data = {'foo': t}
+            create_query = "CREATE TABLE test (id int, value json);"
+            insert_query = """INSERT INTO test (id, value) VALUES (1, '%s');""" % json.dumps(data)
+            event = self.create_and_insert_value(create_query, insert_query)
+            self.assertEqual(event.rows[0]["values"]["value"], to_binary_dict(data))
+
+            self.tearDown()
+            self.setUp()
+
+    def test_json_basic(self):
+        if not self.isMySQL57():
+            self.skipTest("Json is only supported in mysql 5.7")
+
+        types = [
+           True,
+           False,
+           None,
+           1.2,
+           2^14,
+           2^30,
+           2^62,
+           -1 * 2^14,
+           -1 * 2^30,
+           -1 * 2^62,
+        ]
+
+        for data in types:
+            create_query = "CREATE TABLE test (id int, value json);"
+            insert_query = """INSERT INTO test (id, value) VALUES (1, '%s');""" % json.dumps(data)
+            event = self.create_and_insert_value(create_query, insert_query)
+            self.assertEqual(event.rows[0]["values"]["value"], data)
+
+            self.tearDown()
+            self.setUp()
+
+    def test_json_unicode(self):
+        if not self.isMySQL57():
+            self.skipTest("Json is only supported in mysql 5.7")
+        create_query = "CREATE TABLE test (id int, value json);"
+        insert_query = u"""INSERT INTO test (id, value) VALUES (1, '{"miam": "🍔"}');"""
+        event = self.create_and_insert_value(create_query, insert_query)
+        self.assertEqual(event.rows[0]["values"]["value"][b"miam"], u'🍔'.encode('utf8'))
+
+    def test_json_long_string(self):
+        if not self.isMySQL57():
+            self.skipTest("Json is only supported in mysql 5.7")
+        create_query = "CREATE TABLE test (id int, value json);"
+        # The string length needs to be larger than what can fit in a single byte.
+        string_value = "super_long_string" * 100
+        insert_query = "INSERT INTO test (id, value) VALUES (1, '{\"my_key\": \"%s\"}');" % (string_value,)
+        event = self.create_and_insert_value(create_query, insert_query)
+        self.assertEqual(event.rows[0]["values"]["value"], to_binary_dict({"my_key": string_value}))
+
     def test_null(self):
         create_query = "CREATE TABLE test ( \
             test TINYINT NULL DEFAULT NULL, \
